# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.Let's use the full copyright notice
"""This script is to be run by build_dataset.sh.

It turns the data from txt files to a python dictionary,
and then using SQLiteFederatedDataBuilder it moves the dictionary to SQLite.
This file also outputs the normalized_lines text file to be used for the
vocabulary file with most frequent words.
"""

import os
import re
import string

from absl import app
from absl import flags
import fedjax
import numpy as np

FLAGS = flags.FLAGS
flags.DEFINE_string('data_dir', '/tmp/cornell/cornell_dataset',
                    'Path for cornell dataset.')

flags.DEFINE_string(
    'out_dir', '/dev/cornell',
    'Path for the sqlite file and normalized_lines.txt file.')


def main(argv) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')
  federated_dataset_dict = {}
  # Get line id and dialogue data from movie_lines.txt file.
  with open(
      os.path.join(FLAGS.data_dir, 'movie_lines.txt'),
      'r',
      encoding='iso-8859-1') as text:
    for line in text:
      line_id, client_id, _, _, dialogue = line.split(' +++$+++ ')
      if ' CONTINUED:' in dialogue:
        continue
      normalized_string = normalize_dialogue_string(dialogue)
      if not normalized_string or normalized_string.isspace():
        continue
      if client_id not in federated_dataset_dict:
        federated_dataset_dict[client_id] = {
            'line_id': [line_id],
            'line': [normalized_string]
        }
      else:
        federated_dataset_dict[client_id]['line_id'].append(line_id)
        federated_dataset_dict[client_id]['line'].append(normalized_string)

  # Writing normalized lines to file for word count for vocabulary generation.
  with open(os.path.join(FLAGS.out_dir, 'normalized_lines.txt'),
            'w') as file:
    for client_id in federated_dataset_dict:
      for line in federated_dataset_dict[client_id]['line']:
        file.write(line)
        file.write('\n')
  # Get name, movie id, and gender data from movie_characters_metadata.txt file.
  with open(
      os.path.join(FLAGS.data_dir,
                   'movie_characters_metadata.txt'),
      'r',
      encoding='iso-8859-1') as text:
    for line in text:
      client_id, name, movie_id, _, gender, _ = line.split(' +++$+++ ')
      if client_id in federated_dataset_dict:
        num_examples = len(federated_dataset_dict[client_id]['line'])

        federated_dataset_dict[client_id]['name'] = np.full(num_examples, name)
        federated_dataset_dict[client_id]['movie'] = np.full(
            num_examples, movie_id)
        federated_dataset_dict[client_id]['gender'] = np.full(
            num_examples, gender)

  for client_id in federated_dataset_dict:
    federated_dataset_dict[client_id]['line_id'] = np.array(
        federated_dataset_dict[client_id]['line_id'])
    federated_dataset_dict[client_id]['line'] = np.array(
        federated_dataset_dict[client_id]['line'])

  with fedjax.SQLiteFederatedDataBuilder(
      os.path.join(FLAGS.out_dir,
                   'cornell_dataset.sqlite')) as builder:
    builder.add_many(
        (cid.encode(), cds) for cid, cds in federated_dataset_dict.items())


def normalize_dialogue_string(dialogue: str) -> str:
  """Normalizes the lines (dialogue).

  1. Removes newlines, punctuation and decases the string.
  2. Removes ellipses.
  3. Removes double hyphens.
  4. Numeric values are replaced with <NUMERIC>.
  3. Repalces any two or more spaces with one space.

  Args:
    dialogue: line of dialogue from the movie_lines.txt file

  Returns:
    The normalized string
  """
  strip_newline = dialogue.strip('\n')
  no_html_tags = re.sub(r'</?.>', '', strip_newline)
  no_punc_decased = ' '.join(
      [w.strip(string.punctuation) for w in no_html_tags.lower().split(' ')])
  no_ellipses = re.sub(r'(\.+)\1{1,}', ' ', no_punc_decased)
  no_double_hyphen = re.sub(r'--', ' ', no_ellipses)
  no_numbers = re.sub(r' \d+(,\d\d\d)*(\.\d+)?', ' <NUMERIC> ',
                      '' + no_double_hyphen + '').strip()
  return re.sub(r'\s{2,}', ' ', no_numbers)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Writes CSV data to SQLite.

Reads from the all_data.csv file created in preprocess.sh. Since the data in
all_data.csv is not sorted by user, this file sorts the data by user. For each
user, it then gathers the user's tweets, does some minimal preprocessing where
punctuation is removed, all text is lowercased, and features are reduced. The
reduced features are: numbers -> <NUMERIC> @... -> <USERNAME> http... -> <URL>.
The preprocessed tweets are then written to a normalized_lines.txt file, which
contains all preprocessed tweets and aids in the creation of vocab.txt. The
preprocessed tweets, alongside other user metadata, are finally converted to
NumPy arrays and written to SQLite.

It is important to note that on the original data set, the training data only
contains positive and negative labels, 4 and 0, respectively. However, the
testing data contains positive, neutral, and negative labels, 4, 2, and 0,
respectively. To maintain consistency with the leaf repository, this file
converts positive labels (4) to 1 and neutral (2) and negative labels (0) to 0.
"""
import csv
import os
import re
import string

from typing import List, Dict
from absl import app
from absl import flags
import fedjax
import numpy as np

FLAGS = flags.FLAGS

flags.DEFINE_string('out_dir', '/Users',
                    'Path for the sqlite file and normalized_lines.txt file.')

flags.DEFINE_string('data_dir', '/tmp/sent140',
                    'Path for Sent140 data set.')


def main(argv) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  data_file = os.path.join(FLAGS.data_dir, 'all_data.csv')
  data = {}
  raw_data = []
  with open(data_file, 'rt', encoding='ISO-8859-1') as f:
    reader = csv.reader(f)
    raw_data = list(reader)
    for line in raw_data:
      user = line[4]
      if user not in data:
        data[user] = [line]
      else:
        data[user].append(line)

  with fedjax.SQLiteFederatedDataBuilder(
      os.path.join(FLAGS.out_dir, 'sent140_dataset.sqlite')) as builder:
    builder.add_many(
        (cid.encode(), user_data_to_numpy(cd)) for cid, cd in data.items())

  # Writing normalized lines to file for word count for vocabulary generation.
  with open(
      os.path.join(FLAGS.out_dir, 'normalized_lines.txt'),
      'w') as file:
    for row in raw_data:
      file.write(normalize_tweet(row[5]))
      file.write('\n')


def user_data_to_numpy(user_data: List[List[str]]) -> Dict[str, np.ndarray]:
  """Formats a user's data as NumPy arrays.

  Normalizes tweets with normalize_tweet.

  Args:
    user_data: An array containing a user's data.

  Returns:
    A dict[str, np.ndarray]
  """
  tweet_id = []
  tweet_date = []
  tweet_query = []
  tweet_client = []
  tweet_text = []
  tweet_dataset = []
  tweet_sentiment = []

  for row in user_data:
    # Converts positive sentiment label of 4 to 1. Converts neutral and negative
    # sentiment labels of 2 and 0 to 0. Done to maintain homogeneity with leaf
    # repository.
    y = '1' if row[0] == '4' else '0'
    tweet_id.append(bytes(row[1], 'utf-8'))
    tweet_date.append(bytes(row[2], 'utf-8'))
    tweet_query.append(bytes(row[3], 'utf-8'))
    tweet_client.append(bytes(row[4], 'utf-8'))
    tweet_text.append(bytes(normalize_tweet(row[5]), 'utf-8'))
    tweet_dataset.append(bytes(row[6], 'utf-8'))
    tweet_sentiment.append(bytes(y, 'utf-8'))

  np_data = {
      'tweet_id': np.array(tweet_id, dtype=np.object),
      'tweet_date': np.array(tweet_date, dtype=np.object),
      'tweet_query': np.array(tweet_query, dtype=np.object),
      'tweet_client': np.array(tweet_client, dtype=np.object),
      'tweet_text': np.array(tweet_text, dtype=np.object),
      'tweet_dataset': np.array(tweet_dataset, dtype=np.object),
      'tweet_sentiment': np.array(tweet_sentiment, dtype=np.object)
  }

  return np_data


def normalize_tweet(tweet: str) -> str:
  """Normalizes the tweet.

  1. Removes punctuation and decases the tweet.
  2. Usernames (@...) are replaced with <USERNAME>.
  3. Urls (http...) are replaced with <URL>.
  4. Numeric values are replaced with <NUMERIC>.

  Args:
    tweet: single tweet from sent140 data set.

  Returns:
    The normalized tweet
  """
  decased = tweet.lower()
  no_urls = re.sub(r'http\S+', ' URL ', decased)
  no_usernames = re.sub(r'@\S+', ' USERNAME ', no_urls)
  no_numerics = re.sub(r'\d+(,*)\d*(\.*)\d*(-*)\d*', ' NUMERIC ', no_usernames)
  no_punc = ' '.join([
      # strips every punctuation but apostrophes and hyphens
      w.strip(string.punctuation[:6] + string.punctuation[7:12] +
              string.punctuation[13:])
      for w in re.findall(r'\w+|[^\s\w]+', no_numerics)
  ])
  no_double_hyphen = re.sub(r'-{2,}', ' ', no_punc)
  format_apostrophes = re.sub(r" ' ", "'", no_double_hyphen)
  format_hyphens = re.sub(r' - ', '-', format_apostrophes)
  single_spaces = ' '.join(format_hyphens.split())

  return single_spaces


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.Let's use the full copyright notice
"""Test for data_to_sqlite normalization."""

from absl.testing import absltest
from absl.testing import parameterized
from fedjax.datasets.scripts.sent140 import data_to_sqlite


class DataToSQLiteTest(parameterized.TestCase):

  @parameterized.named_parameters(
      {
          'testcase_name': 'test url and username',
          'input_string': 'Thank @user1 https://en.wikipedia.org more story...',
          'desired': 'thank USERNAME URL more story',
      }, {
          'testcase_name': 'test numerics',
          'input_string': 'Pi 3.14, 10 months, 50,000.00 and 7-10.',
          'desired': 'pi NUMERIC NUMERIC months NUMERIC and NUMERIC',
      }, {
          'testcase_name': 'test hyphen',
          'input_string': 'This is a double--hyphen! And this is a single-one.',
          'desired': 'this is a double hyphen and this is a single-one',
      }, {
          'testcase_name': 'test punctuation',
          'input_string': 'I.do{not},know....that*',
          'desired': 'i do not know that',
      })
  def test_normalize_dialogue_string(self, input_string, desired):
    self.assertEqual(desired, data_to_sqlite.normalize_tweet(input_string))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated hypothesis-based clustering implementation.

Three Approaches for Personalization with Applications to Federated Learning
    Yishay Mansour, Mehryar Mohri, Jae Ro, Ananda Theertha Suresh
    https://arxiv.org/abs/2002.10619
"""
import functools
from typing import Any, Callable, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import models
from fedjax.core import optimizers
from fedjax.core import tree_util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import OptState
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey

import jax
import jax.numpy as jnp


@dataclasses.dataclass
class ServerState:
  """State of server for HypCluster passed between rounds.

  Attributes:
    cluster_params: A list of pytrees representing the server model parameters
      per cluster.
    opt_states: A list of pytrees representing the server optimizer state per
      cluster.
  """
  cluster_params: List[Params]
  opt_states: List[OptState]


def hyp_cluster(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    client_optimizer: optimizers.Optimizer,
    server_optimizer: optimizers.Optimizer,
    maximization_batch_hparams: client_datasets.PaddedBatchHParams,
    expectation_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None
) -> federated_algorithm.FederatedAlgorithm:
  """Federated hypothesis-based clustering algorithm.

  Args:
    per_example_loss: A function from (params, batch, rng) to a vector of per
      example loss values.
    client_optimizer: Client side optimizer.
    server_optimizer: Server side optimizer.
    maximization_batch_hparams: Batching hyperparameters for the maximization
      step.
    expectation_batch_hparams: Batching hyperparameters for the expectation
      step.
    regularizer: Optional regularizer.

  Returns:
    A FederatedAlgorithm. A notable difference from other common
    FederatedAlgorithms such as fed_avg is that `init()` takes a list of
    `cluster_params`, which can be obtained using `random_init()`,
    `ModelKMeansInitializer`, or `kmeans_init()` in this module.
  """

  def init(cluster_params: List[Params]) -> ServerState:
    return ServerState(
        cluster_params,
        [server_optimizer.init(params) for params in cluster_params])

  # Creating these objects outside apply() can speed up repeated apply() calls.
  evaluator = models.AverageLossEvaluator(per_example_loss, regularizer)
  trainer = ClientDeltaTrainer(
      models.grad(per_example_loss, regularizer), client_optimizer)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]]
  ) -> Tuple[ServerState, Mapping[federated_data.ClientId, Any]]:
    # Split RNGs for the 2 steps below.
    client_rngs = [jax.random.split(rng) for _, _, rng in clients]

    client_cluster_ids = maximization_step(
        evaluator=evaluator,
        cluster_params=server_state.cluster_params,
        clients=[(client_id, dataset, rng[0])
                 for (client_id, dataset, _), rng in zip(clients, client_rngs)],
        batch_hparams=maximization_batch_hparams)

    cluster_delta_params = expectation_step(
        trainer=trainer,
        cluster_params=server_state.cluster_params,
        client_cluster_ids=client_cluster_ids,
        clients=[(client_id, dataset, rng[1])
                 for (client_id, dataset, _), rng in zip(clients, client_rngs)],
        batch_hparams=expectation_batch_hparams)

    # Apply delta params for each cluster.
    cluster_params = []
    opt_states = []
    for delta_params, opt_state, params in zip(cluster_delta_params,
                                               server_state.opt_states,
                                               server_state.cluster_params):
      if delta_params is None:
        # No examples were observed for this cluster.
        next_opt_state, next_params = opt_state, params
      else:
        next_opt_state, next_params = server_optimizer.apply(
            delta_params, opt_state, params)
      cluster_params.append(next_params)
      opt_states.append(next_opt_state)

    # TODO(wuke): Other client diagnostics.
    client_diagnostics = {}
    for client_id, cluster_id in client_cluster_ids.items():
      client_diagnostics[client_id] = {'cluster_id': cluster_id}
    return ServerState(cluster_params, opt_states), client_diagnostics

  return federated_algorithm.FederatedAlgorithm(init, apply)


class _BaseClientTrainer:
  """Trains each client dataset, starting from either global params, or per client params."""

  def __init__(self, grad_fn: Callable[[Params, BatchExample, PRNGKey],
                                       jnp.ndarray],
               client_optimizer: optimizers.Optimizer, return_delta):
    # params can be passed in 2 ways:
    # -   As `shared_input`: All clients are trained starting from the same
    #     params.
    # -   As `client_input`: Each client is trained starting from per client
    #     params.
    def client_init(shared_input, client_input):
      if shared_input is not None:
        init_params = shared_input
        rng = client_input
      else:
        rng, init_params = client_input
      params = init_params
      opt_state = client_optimizer.init(params)
      return rng, params, opt_state, init_params

    def client_step(state, batch):
      rng, params, opt_state, init_params = state
      rng, use_rng = jax.random.split(rng)
      grads = grad_fn(params, batch, use_rng)
      opt_state, params = client_optimizer.apply(grads, opt_state, params)
      return rng, params, opt_state, init_params

    def client_final(shared_input, state):
      del shared_input
      _, params, _, init_params = state
      if return_delta:
        delta_params = jax.tree_util.tree_multimap(jnp.subtract, init_params,
                                                   params)
        return delta_params
      return params

    self._train_each_client = for_each_client.for_each_client(
        client_init, client_step, client_final)
    self._return_delta = return_delta

  def returns_delta(self):
    return self._return_delta

  def train_global_params(
      self, params: Params, clients: Iterable[Tuple[federated_data.ClientId,
                                                    Iterable[BatchExample],
                                                    PRNGKey]]
  ) -> Iterator[Tuple[federated_data.ClientId, Params]]:
    yield from self._train_each_client(shared_input=params, clients=clients)

  def train_per_client_params(
      self, clients: Iterable[Tuple[federated_data.ClientId,
                                    Iterable[BatchExample], PRNGKey, Params]]
  ) -> Iterator[Tuple[federated_data.ClientId, Params]]:
    yield from self._train_each_client(
        shared_input=None,
        clients=[(client_id, batches, (rng, params))
                 for client_id, batches, rng, params in clients])


class ClientDeltaTrainer(_BaseClientTrainer):
  """Trains each client dataset, starting from either global params, or per client params.

  The result values are delta params (i.e. initial params - final params).
  """

  def __init__(self, grad_fn: Callable[[Params, BatchExample, PRNGKey],
                                       jnp.ndarray],
               client_optimizer: optimizers.Optimizer):
    super().__init__(grad_fn, client_optimizer, return_delta=True)


class ClientParamsTrainer(_BaseClientTrainer):
  """Trains each client dataset, starting from either global params, or per client params.

  The result values are final params.
  """

  def __init__(self, grad_fn: Callable[[Params, BatchExample, PRNGKey],
                                       jnp.ndarray],
               client_optimizer: optimizers.Optimizer):
    super().__init__(grad_fn, client_optimizer, return_delta=False)


def maximization_step(
    evaluator: models.AverageLossEvaluator, cluster_params: List[Params],
    clients: Sequence[Tuple[federated_data.ClientId,
                            client_datasets.ClientDataset, PRNGKey]],
    batch_hparams: client_datasets.PaddedBatchHParams
) -> Dict[federated_data.ClientId, int]:
  """Evaluates and assigns each client to the cluster with the lowest average loss."""
  cluster_losses = _cluster_losses(
      evaluator=evaluator,
      cluster_params=cluster_params,
      clients=clients,
      batch_hparams=batch_hparams)
  return jax.device_get(_cluster_assignment(cluster_losses))


def _cluster_losses(
    evaluator: models.AverageLossEvaluator, cluster_params: List[Params],
    clients: Sequence[Tuple[federated_data.ClientId,
                            client_datasets.ClientDataset, PRNGKey]],
    batch_hparams: client_datasets.PaddedBatchHParams
) -> Dict[federated_data.ClientId, List[jnp.ndarray]]:
  """Evaluates the average loss for each client and cluster."""
  cluster_losses = {client_id: [] for client_id, _, _ in clients}
  client_rngs = [
      jax.random.split(rng, len(cluster_params)) for _, _, rng in clients
  ]
  for i, params in enumerate(cluster_params):
    for client_id, average_loss in evaluator.evaluate_global_params(
        params,
        [(client_id, dataset.padded_batch(batch_hparams), rng[i])
         for (client_id, dataset, _), rng in zip(clients, client_rngs)]):
      cluster_losses[client_id].append(average_loss)
  return cluster_losses


@jax.jit
def _cluster_assignment(cluster_losses):
  """Assigns each client the cluster with the lowest loss."""
  return {
      client_id: jnp.argmin(jnp.stack(losses))
      for client_id, losses in cluster_losses.items()
  }


def expectation_step(trainer: ClientDeltaTrainer, cluster_params: List[Params],
                     client_cluster_ids: Dict[federated_data.ClientId, int],
                     clients: Sequence[Tuple[federated_data.ClientId,
                                             client_datasets.ClientDataset,
                                             PRNGKey]],
                     batch_hparams: client_datasets.ShuffleRepeatBatchHParams):
  """Updates each cluster's params using average delta from clients in this cluster."""
  # Train each client starting from its corresponding cluster params, and
  # calculate weighted average of delta params within each cluster.
  num_examples = {client_id: len(dataset) for client_id, dataset, _ in clients}
  cluster_delta_params_sum = [
      jax.tree_util.tree_map(jnp.zeros_like, params)
      for params in cluster_params
  ]
  cluster_num_examples_sum = [0 for _ in cluster_params]
  for client_id, delta_params in trainer.train_per_client_params([
      (client_id, dataset.shuffle_repeat_batch(batch_hparams), rng,
       cluster_params[client_cluster_ids[client_id]])
      for client_id, dataset, rng in clients
  ]):
    cluster_id = client_cluster_ids[client_id]
    cluster_delta_params_sum[cluster_id] = tree_util.tree_add(
        cluster_delta_params_sum[cluster_id],
        tree_util.tree_weight(delta_params, num_examples[client_id]))
    cluster_num_examples_sum[cluster_id] += num_examples[client_id]
  # Weighted average delta params, or None if no examples were seen for this
  # cluster.
  cluster_delta_params = []
  for delta_params_sum, num_examples_sum in zip(cluster_delta_params_sum,
                                                cluster_num_examples_sum):
    if num_examples_sum > 0:
      cluster_delta_params.append(
          tree_util.tree_inverse_weight(delta_params_sum, num_examples_sum))
    else:
      cluster_delta_params.append(None)
  return cluster_delta_params


@functools.partial(jax.jit, static_argnums=(0, 1))
def random_init(num_clusters: int, init: Callable[[PRNGKey], Params],
                rng: PRNGKey) -> List[Params]:
  """Randomly initializes cluster params."""
  return [init(i) for i in jax.random.split(rng, num_clusters)]


class ModelKMeansInitializer:
  """Initializes cluster params for HypCluster using a k-means++ variant.

  This is a thin wrapper for initializing from a
  :class:`Model <fedjax.core.models.Model>` .
  See :func:`kmeans_init` for a more general version of the initializer, and
  details of the initialization algorithm.
  """

  def __init__(self,
               model: models.Model,
               client_optimizer: optimizers.Optimizer,
               regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
    self._model = model
    self._trainer = ClientParamsTrainer(
        models.model_grad(model, regularizer), client_optimizer)
    self._evaluator = models.AverageLossEvaluator(
        models.model_per_example_loss(model), regularizer)

  def cluster_params(
      self, num_clusters: int, rng: PRNGKey,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]],
      train_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
      eval_batch_hparams: client_datasets.PaddedBatchHParams) -> List[Params]:
    # Shared initial params.
    init_rng, center_rng = jax.random.split(rng)
    init_params = self._model.init(init_rng)
    return kmeans_init(
        num_clusters=num_clusters,
        init_params=init_params,
        clients=clients,
        trainer=self._trainer,
        train_batch_hparams=train_batch_hparams,
        evaluator=self._evaluator,
        eval_batch_hparams=eval_batch_hparams,
        rng=center_rng)


def kmeans_init(num_clusters: int, init_params: Params,
                clients: Sequence[Tuple[federated_data.ClientId,
                                        client_datasets.ClientDataset,
                                        PRNGKey]], trainer: ClientParamsTrainer,
                train_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
                evaluator: models.AverageLossEvaluator,
                eval_batch_hparams: client_datasets.PaddedBatchHParams,
                rng: PRNGKey) -> List[Params]:
  """Initializes cluster params for HypCluster using a k-means++ variant.

  See :class:`ModelKMeansInitializer` for a more convenient initializer from a
  :class:`Model <fedjax.core.models.Model>`.

  Given a set of input clients, we train parameters for each client. The
  initial cluster parameters are chosen out of this set of client parameters.
  At a high level, the cluster parameters are selected by choosing the client
  parameters that are the "furthest away" from each other (greatest difference
  in loss).

  Args:
    num_clusters: Desired number of cluster centers.
    init_params: Initial params for training each client.
    clients: Clients to train for generating candidate cluster center params.
    trainer: Client trainer for carry out client training.
    train_batch_hparams: Batching hyperparameters for client training.
    evaluator: Average loss evaluator for evaluator clients on chosen cluster
      center params.
    eval_batch_hparams: Batching hyperparameters for client evaluation.
    rng: RNG used for choosing the initial cluster center.

  Returns:
    Cluster center params.

  Raises:
    ValueError if some input arguments are invalid.
  """
  if num_clusters <= 0:
    raise ValueError(f'num_clusters should be > 0, got {num_clusters}')
  # Choose a client to be the initial center.
  initial_center_index = jax.random.choice(rng, len(clients))
  del rng  # Free up the name "rng" for subsequent loops.

  # Split client RNGs. For each client, we need 1 RNG for training,
  # (num_clusters - 1) for average loss evaluation.
  client_rngs = [jax.random.split(rng, num_clusters) for _, _, rng in clients]

  # Train each client from init_params.
  train_clients = [
      (client_id, dataset.shuffle_repeat_batch(train_batch_hparams), rng[0])
      for (client_id, dataset, _), rng in zip(clients, client_rngs)
  ]
  client_params = dict(trainer.train_global_params(init_params, train_clients))

  # TODO(wuke): Should clients already chosen be eliminated from subsequent
  # evaluations to prevent the same center from being added multiple times?

  # Choose cluster centers, starting with a randomly chosen initial center.
  cluster_centers = [client_params[sorted(client_params)[initial_center_index]]]
  # Repeatedly add the client with the worse best loss as the next center.
  best_losses = {client_id: jnp.inf for client_id in client_params}
  for i in range(1, num_clusters):
    # Evaluate all clients with the newly added center.
    eval_clients = [
        (client_id, dataset.padded_batch(eval_batch_hparams), rng[i])
        for (client_id, dataset, _), rng in zip(clients, client_rngs)
    ]
    new_losses = dict(
        evaluator.evaluate_global_params(cluster_centers[-1], eval_clients))
    # Update best loss for each client.
    best_losses = jax.tree_util.tree_multimap(min, best_losses,
                                              jax.device_get(new_losses))
    # Add a new center.
    worst_client = max(best_losses, key=lambda k: best_losses[k])
    cluster_centers.append(client_params[worst_client])

  return cluster_centers


class HypClusterEvaluator:
  """Evaluates cluster params on a Model."""

  def __init__(self,
               model: models.Model,
               regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
    """Initializes some reusable components.

    Because we need to make multiple passes over each client during evaluation,
    the number of clients that can be evaluated at once is limited. Therefore
    multiple calls to :meth:`~HypClusterEvaluator.evaluate_clients` are needed
    to evaluate a large federated dataset. We factor out some reusable
    components so that the same computation can be jit compiled.

    Args:
      model: Model being evaluated.
      regularizer: Optional regularizer.
    """
    self._maximization_step_evaluator = models.AverageLossEvaluator(
        models.model_per_example_loss(model), regularizer)
    self._model_evaluator = models.ModelEvaluator(model)

  def evaluate_clients(
      self, cluster_params: List[Params],
      train_clients: Sequence[Tuple[federated_data.ClientId,
                                    client_datasets.ClientDataset, PRNGKey]],
      test_clients: Sequence[Tuple[federated_data.ClientId,
                                   client_datasets.ClientDataset]],
      batch_hparams: client_datasets.PaddedBatchHParams
  ) -> Iterator[Tuple[federated_data.ClientId, Dict[str, jnp.ndarray]]]:
    """Evaluates each client on the cluster with best average loss."""
    cluster_client_ids = maximization_step(
        evaluator=self._maximization_step_evaluator,
        cluster_params=cluster_params,
        clients=train_clients,
        batch_hparams=batch_hparams)
    eval_clients = [(client_id, dataset.padded_batch(batch_hparams),
                     cluster_params[cluster_client_ids[client_id]])
                    for client_id, dataset in test_clients]
    yield from self._model_evaluator.evaluate_per_client_params(eval_clients)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CIFAR-100 models."""

from fedjax.core import metrics
from fedjax.core import models

import haiku as hk
import numpy as np

# Defines the expected structure of input batches to the model. This is used to
# determine the model parameter shapes.
_HAIKU_SAMPLE_BATCH = {
    'x': np.zeros((1, 24, 24, 3), dtype=np.float32),
    'y': np.zeros(1, dtype=np.int32)
}
_TRAIN_LOSS = lambda batch, params: metrics.unreduced_cross_entropy_loss(
    batch['y'], params)
_EVAL_METRICS = {
    'loss': metrics.CrossEntropyLoss(),
    'accuracy': metrics.Accuracy(),
}


def create_logistic_model() -> models.Model:
  """Creates CIFAR-100 logistic model with haiku."""
  num_classes = 100

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Small util functions with no better place to go."""

import jax.numpy as jnp


# Background: https://jax.readthedocs.io/en/latest/faq.html#gradients-contain-nan-where-using-where
def safe_div(a, b):
  """Divides a by b, or returns 0 if b is 0 in a safe manner for gradient computation."""
  safe = b != 0
  c = a / jnp.where(safe, b, 1)
  return jnp.where(safe, c, 0)


class FakeTensorFlowModule:

  def __getattr__(self, _):
    raise ModuleNotFoundError(
        """This FedJAX feature requires TensorFlow, but TensorFlow is not installed.

If you do not otherwise need TensorFlow, we recommend installing the smaller
CPU-only version of TensorFlow via

  pip install tensorflow-cpu

If you may need to use TensorFlow with GPU, please install the full version via

  pip install tensorflow
""")


# TODO(b/188556866): Remove dependency on TensorFlow.
def import_tf():
  """Imports and returns TensorFlow module if it is installed.

  This is to avoid having FedJAX directly depend on TensorFlow since TensorFlow
  is large and complex and the majority of FedJAX functions without it.

  TensorFlow must be imported before calling absl.main() or absltest.main() to
  ensure proper TensorFlow initialization. Examples:

  * Do this::

      # At the top of a Python file along with other imports.
      import numpy as np
      from fedjax.core import util
      tf = util.import_tf()

  * Don't do this::

      # Inside a function (i.e. delayed import).
      def my_tf_function():
        from fedjax.core import util
        tf = util.import_tf()

  Returns:
    TensorFlow module if it is installed. Otherwise, an object that serves as an
    fake module that throws an exception when accessed.
  """
  try:
    import tensorflow  # pylint:disable=g-import-not-at-top
    return tensorflow
  except ModuleNotFoundError:
    return FakeTensorFlowModule()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.models."""

from absl.testing import absltest

from fedjax.core import metrics
from fedjax.core import models

import haiku as hk
import jax
try:
  from jax.example_libraries import stax
except ModuleNotFoundError:
  from jax.experimental import stax
import jax.numpy as jnp
import numpy as np
import numpy.testing as npt

train_loss = lambda b, p: metrics.unreduced_cross_entropy_loss(b['y'], p)
eval_metrics = {'accuracy': metrics.Accuracy()}


class ModelTest(absltest.TestCase):

  def check_model(self, model):
    with self.subTest('init'):
      params = model.init(jax.random.PRNGKey(0))
      num_params = sum(l.size for l in jax.tree_util.tree_leaves(params))
      self.assertEqual(num_params, 30)

    with self.subTest('apply_for_train'):
      batch = {
          'x': np.array([[1, 2], [3, 4], [5, 6]]),
          'y': np.array([7, 8, 9])
      }
      preds = model.apply_for_train(params, batch, jax.random.PRNGKey(0))
      self.assertTupleEqual(preds.shape, (3, 10))

    with self.subTest('apply_for_eval'):
      preds = model.apply_for_eval(params, batch)
      self.assertTupleEqual(preds.shape, (3, 10))

    with self.subTest('train_loss'):
      preds = model.apply_for_train(params, batch, jax.random.PRNGKey(0))
      loss = model.train_loss(batch, preds)
      self.assertTupleEqual(loss.shape, (3,))

  def test_create_model_from_haiku(self):

    def forward_pass(batch):
      return hk.Linear(10)(batch['x'])

    haiku_model = models.create_model_from_haiku(
        transformed_forward_pass=hk.transform(forward_pass),
        sample_batch={'x': jnp.ones((1, 2))},
        train_loss=train_loss,
        eval_metrics=eval_metrics)
    self.check_model(haiku_model)

  def test_create_model_from_stax(self):
    stax_init, stax_apply = stax.serial(stax.Dense(10))
    stax_model = models.create_model_from_stax(
        stax_init=stax_init,
        stax_apply=stax_apply,
        sample_shape=(-1, 2),
        train_loss=train_loss,
        eval_metrics=eval_metrics)
    self.check_model(stax_model)

  def test_evaluate_model(self):
    # Mock out Model.
    model = models.Model(
        init=lambda rng: None,  # Unused.
        apply_for_train=lambda params, batch, rng: None,  # Unused.
        apply_for_eval=lambda params, batch: batch.get('pred'),
        train_loss=lambda batch, preds: None,  # Unused.
        eval_metrics={
            'accuracy': metrics.Accuracy(),
            'loss': metrics.CrossEntropyLoss(),
        })
    params = jnp.array(3.14)  # Unused.
    batches = [
        {
            'y': np.array([1, 0, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
        },
        {
            'y': np.array([0, 1, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
        },
        {
            'y': np.array([0, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1]])
        },
    ]
    eval_results = models.evaluate_model(model, params, batches)
    self.assertEqual(eval_results['accuracy'], 0.625)  # 5 / 8.
    self.assertAlmostEqual(eval_results['loss'], 0.8419596, places=6)

  def test_evaluate_global_params(self):
    # Mock out Model.
    model = models.Model(
        init=lambda rng: None,  # Unused.
        apply_for_train=lambda params, batch, rng: None,  # Unused.
        apply_for_eval=lambda params, batch: batch.get('pred') + params,
        train_loss=lambda batch, preds: None,  # Unused.
        eval_metrics={
            'accuracy': metrics.Accuracy(),
            'loss': metrics.CrossEntropyLoss(),
        })
    params = jnp.array([1, -1])
    clients = [
        (b'0000', [{
            'y': np.array([1, 0, 1]),
            'pred': np.array([[0.2, 1.4], [1.3, 1.1], [-0.7, 4.2]])
        }, {
            'y': np.array([0, 1, 1]),
            'pred': np.array([[0.2, 1.4], [1.3, 1.1], [-0.7, 4.2]])
        }]),
        (b'1001', [{
            'y': np.array([0, 1]),
            'pred': np.array([[0.2, 1.4], [1.3, 1.1]])
        }]),
    ]
    eval_results = dict(
        models.ModelEvaluator(model).evaluate_global_params(params, clients))
    self.assertCountEqual(eval_results, [b'0000', b'1001'])
    self.assertCountEqual(eval_results[b'0000'], ['accuracy', 'loss'])
    self.assertAlmostEqual(eval_results[b'0000']['accuracy'], 4 / 6)
    self.assertAlmostEqual(eval_results[b'0000']['loss'], 0.67658216)
    self.assertCountEqual(eval_results[b'1001'], ['accuracy', 'loss'])
    self.assertEqual(eval_results[b'1001']['accuracy'], 1 / 2)
    self.assertAlmostEqual(eval_results[b'1001']['loss'], 1.338092, places=6)

  def test_evaluate_per_client_params(self):
    # Mock out Model.
    model = models.Model(
        init=lambda rng: None,  # Unused.
        apply_for_train=lambda params, batch, rng: None,  # Unused.
        apply_for_eval=lambda params, batch: batch.get('pred') + params,
        train_loss=lambda batch, preds: None,  # Unused.
        eval_metrics={
            'accuracy': metrics.Accuracy(),
            'loss': metrics.CrossEntropyLoss(),
        })
    clients = [
        (b'0000', [{
            'y': np.array([1, 0, 1]),
            'pred': np.array([[0.2, 1.4], [1.3, 1.1], [-0.7, 4.2]])
        }, {
            'y': np.array([0, 1, 1]),
            'pred': np.array([[0.2, 1.4], [1.3, 1.1], [-0.7, 4.2]])
        }], jnp.array([1, -1])),
        (b'1001', [{
            'y': np.array([0, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1]])
        }], jnp.array([0, 0])),
    ]
    eval_results = dict(
        models.ModelEvaluator(model).evaluate_per_client_params(clients))
    self.assertCountEqual(eval_results, [b'0000', b'1001'])
    self.assertCountEqual(eval_results[b'0000'], ['accuracy', 'loss'])
    npt.assert_allclose(eval_results[b'0000']['accuracy'], 4 / 6)
    npt.assert_allclose(eval_results[b'0000']['loss'], 0.67658216)
    self.assertCountEqual(eval_results[b'1001'], ['accuracy', 'loss'])
    npt.assert_allclose(eval_results[b'1001']['accuracy'], 1 / 2)
    npt.assert_allclose(eval_results[b'1001']['loss'], 1.338092)

  def test_model_per_example_loss(self):
    # Mock out Model.
    model = models.Model(
        init=lambda rng: None,  # Unused.
        apply_for_train=lambda params, batch, rng: batch['x'] * params + rng,
        apply_for_eval=lambda params, batch: None,  # Unused.
        train_loss=lambda batch, preds: jnp.abs(batch['y'] - preds),
        eval_metrics={}  # Unused
    )

    params = jnp.array(2.)
    batch = {'x': jnp.array([1., -1., 1.]), 'y': jnp.array([0.1, -0.1, -0.1])}
    rng = jnp.array(0.5)

    loss = models.model_per_example_loss(model)(params, batch, rng)
    npt.assert_allclose(loss, [2.4, 1.4, 2.6])

  def test_model_grad(self):
    # Mock out Model.
    model = models.Model(
        init=lambda rng: None,  # Unused.
        apply_for_train=lambda params, batch, rng: batch['x'] * params + rng,
        apply_for_eval=lambda params, batch: None,  # Unused.
        train_loss=lambda batch, preds: jnp.square(batch['y'] - preds) / 2,
        eval_metrics={}  # Unused
    )

    params = jnp.array(2.)
    batch = {'x': jnp.array([1., -1., 1.]), 'y': jnp.array([0.1, -0.1, -0.1])}
    rng = jnp.array(0.5)

    with self.subTest('no regularizer'):
      grads = models.model_grad(model)(params, batch, rng)
      npt.assert_allclose(grads, (2.4 + 1.4 + 2.6) / 3)

    with self.subTest('has regularizer'):
      grads = models.model_grad(model, jnp.abs)(params, batch, rng)
      npt.assert_allclose(grads, (2.4 + 1.4 + 2.6) / 3 + 1)

    with self.subTest('has mask'):
      grads = models.model_grad(model)(params, {
          **batch, '__mask__': jnp.array([True, False, True])
      }, rng)
      npt.assert_allclose(grads, (2.4 + 2.6) / 2)

    with self.subTest('has regularizer, has mask'):
      grads = models.model_grad(model, jnp.abs)(params, {
          **batch, '__mask__': jnp.array([True, False, True])
      }, rng)
      npt.assert_allclose(grads, (2.4 + 2.6) / 2 + 1)


class AverageLossTest(absltest.TestCase):

  def test_evaluate_average_loss(self):

    def per_example_loss(params, batch, rng):
      return params + batch['x'] + jax.random.uniform(rng, [])

    def regularizer(params):
      return 0.5 * jnp.sum(jnp.square(params))

    params = jnp.array(1)

    rng = jax.random.PRNGKey(0)
    rng_uniform_0 = jax.random.uniform(jax.random.split(rng)[1], [])
    rng_uniform_1 = jax.random.uniform(
        jax.random.split(jax.random.split(rng)[0])[1], [])
    rng_term = (rng_uniform_0 * 2 + rng_uniform_1 * 3) / 5

    with self.subTest('no mask, no regularizer'):
      batches = [{'x': jnp.array([1, 2])}, {'x': jnp.array([3, 4, 5])}]
      average_loss = models.evaluate_average_loss(
          params=params,
          batches=batches,
          rng=rng,
          per_example_loss=per_example_loss)
      npt.assert_allclose(average_loss, 4 + rng_term, rtol=1e-6)

    with self.subTest('no mask, has regularizer'):
      batches = [{'x': jnp.array([1, 2])}, {'x': jnp.array([3, 4, 5])}]
      average_loss = models.evaluate_average_loss(
          params=params,
          batches=batches,
          rng=rng,
          per_example_loss=per_example_loss,
          regularizer=regularizer)
      npt.assert_allclose(average_loss, 4.5 + rng_term, rtol=1e-6)

    with self.subTest('has mask, no regularizer'):
      batches = [{
          'x': jnp.array([1, 2, 10]),
          '__mask__': jnp.array([True, True, False])
      }, {
          'x': jnp.array([3, 4, 5])
      }]
      average_loss = models.evaluate_average_loss(
          params=params,
          batches=batches,
          rng=rng,
          per_example_loss=per_example_loss)
      npt.assert_allclose(average_loss, 4 + rng_term, rtol=1e-6)

  def test_evaluate_global_params(self):

    def per_example_loss(params, batch, rng):
      return params + batch['x'] + jax.random.uniform(rng, [])

    def regularizer(params):
      return 0.5 * jnp.sum(jnp.square(params))

    params = jnp.array(1)

    rng_0 = jax.random.PRNGKey(0)
    rng_uniform_00 = jax.random.uniform(jax.random.split(rng_0)[1], [])
    rng_term_0 = rng_uniform_00
    rng_1 = jax.random.PRNGKey(1)
    rng_uniform_10 = jax.random.uniform(jax.random.split(rng_1)[1], [])
    rng_uniform_11 = jax.random.uniform(
        jax.random.split(jax.random.split(rng_1)[0])[1], [])
    rng_term_1 = (rng_uniform_10 * 2 + rng_uniform_11) / 3

    with self.subTest('no mask, no regularizer'):
      clients = [
          (b'0000', [{
              'x': jnp.array([1, 2])
          }], rng_0),
          (b'1001', [{
              'x': jnp.array([3, 4])
          }, {
              'x': jnp.array([5])
          }], rng_1),
      ]
      average_loss = dict(
          models.AverageLossEvaluator(per_example_loss).evaluate_global_params(
              params=params, clients=clients))
      npt.assert_equal(average_loss, {
          b'0000': np.array(2.5) + rng_term_0,
          b'1001': np.array(5) + rng_term_1
      })

    with self.subTest('no mask, has regularizer'):
      clients = [
          (b'0000', [{
              'x': jnp.array([1, 2])
          }], rng_0),
          (b'1001', [{
              'x': jnp.array([3, 4])
          }, {
              'x': jnp.array([5])
          }], rng_1),
      ]
      average_loss = dict(
          models.AverageLossEvaluator(per_example_loss,
                                      regularizer).evaluate_global_params(
                                          params=params, clients=clients))
      npt.assert_equal(average_loss, {
          b'0000': np.array(3) + rng_term_0,
          b'1001': np.array(5.5) + rng_term_1
      })

    with self.subTest('has mask, no regularizer'):
      clients = [
          (b'0000', [{
              'x': jnp.array([1, 2])
          }], rng_0),
          (b'1001', [{
              'x': jnp.array([3, 4])
          }, {
              'x': jnp.array([5, 10]),
              '__mask__': jnp.array([True, False])
          }], rng_1),
      ]
      average_loss = dict(
          models.AverageLossEvaluator(per_example_loss).evaluate_global_params(
              params=params, clients=clients))
      npt.assert_equal(average_loss, {
          b'0000': np.array(2.5) + rng_term_0,
          b'1001': np.array(5) + rng_term_1
      })

  def test_evaluate_per_client_params(self):

    def per_example_loss(params, batch, rng):
      return params + batch['x'] + jax.random.uniform(rng, [])

    def regularizer(params):
      return 0.5 * jnp.sum(jnp.square(params))

    rng_0 = jax.random.PRNGKey(0)
    rng_uniform_00 = jax.random.uniform(jax.random.split(rng_0)[1], [])
    rng_term_0 = rng_uniform_00
    rng_1 = jax.random.PRNGKey(1)
    rng_uniform_10 = jax.random.uniform(jax.random.split(rng_1)[1], [])
    rng_uniform_11 = jax.random.uniform(
        jax.random.split(jax.random.split(rng_1)[0])[1], [])
    rng_term_1 = (rng_uniform_10 * 2 + rng_uniform_11) / 3

    with self.subTest('no mask, no regularizer'):
      clients = [
          (b'0000', [{
              'x': jnp.array([2, 3])
          }], rng_0, jnp.array(0)),
          (b'1001', [{
              'x': jnp.array([3, 4])
          }, {
              'x': jnp.array([5])
          }], rng_1, jnp.array(1)),
      ]
      average_loss = dict(
          models.AverageLossEvaluator(
              per_example_loss).evaluate_per_client_params(clients=clients))
      npt.assert_equal(average_loss, {
          b'0000': np.array(2.5) + rng_term_0,
          b'1001': np.array(5) + rng_term_1
      })

    with self.subTest('no mask, has regularizer'):
      clients = [
          (b'0000', [{
              'x': jnp.array([2, 3])
          }], rng_0, jnp.array(0)),
          (b'1001', [{
              'x': jnp.array([3, 4])
          }, {
              'x': jnp.array([5])
          }], rng_1, jnp.array(1)),
      ]
      average_loss = dict(
          models.AverageLossEvaluator(
              per_example_loss,
              regularizer).evaluate_per_client_params(clients=clients))
      npt.assert_equal(average_loss, {
          b'0000': np.array(2.5) + rng_term_0,
          b'1001': np.array(5.5) + rng_term_1
      })

    with self.subTest('has mask, no regularizer'):
      clients = [
          (b'0000', [{
              'x': jnp.array([2, 3])
          }], rng_0, jnp.array(0)),
          (b'1001', [{
              'x': jnp.array([3, 4])
          }, {
              'x': jnp.array([5, 10]),
              '__mask__': jnp.array([True, False])
          }], rng_1, jnp.array(1)),
      ]
      average_loss = dict(
          models.AverageLossEvaluator(
              per_example_loss).evaluate_per_client_params(clients=clients))
      npt.assert_equal(average_loss, {
          b'0000': np.array(2.5) + rng_term_0,
          b'1001': np.array(5) + rng_term_1
      })


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Efficient Walsh-Hadamard transform in JAX."""

import math
from typing import Tuple, Union

import jax
import jax.numpy as jnp
import scipy
from fedjax.core.typing import PRNGKey, Params


@jax.jit
def walsh_hadamard_transform(
    x: jnp.ndarray,
    small_n: int = 2**7,
    precision: Union[jax.lax.Precision, str] = 'highest') -> jnp.ndarray:
  """Efficient Walsh-Hadamard transform in JAX.

  An accelerator friendly O(n log n) Walsh-Hadamard transform.

  Args:
    x: A vector. len(x) must be a power of 2.
    small_n: Size to break x into. The default value is tuned on TPUv3. Must be
      a power of 2 and > 1.
    precision: Precision for general dot products.

  Returns:
    Transformed vector.
  """
  if small_n <= 1:
    raise ValueError(f'small_n must be > 1, got {small_n}')

  # Let
  # -   A ⊗ B be the Kronecker product of A and B;
  # -   flat(X) be the vector obtained by flattening the rows of X of shape
  #     [M, N].
  #
  # We can show the following:
  #
  #     (A ⊗ B^T) flat(X) = flat(A X B)
  #
  # Note that the Hadamard matrix H_{2^M 2^N} = H_{2^M} ⊗ H_{2^N}, and
  # Hadamard matrices are symmetrical. Therefore, for a [2^M, 2^N] matrix X,
  #
  #     H_{2^M 2^N} flat(X) = flat(H_{2^M} X H_{2^N})
  #
  # The idea can be generalized by breaking a Hadamard matrix into the Kronecker
  # product of many small Hadamard matrices, and reshaping the vector input into
  # a many-dimensional array, and running einsum on each dimension.
  #
  # Let the input vector be of length D, because our "small" Hadamard matrices
  # are of size at most small_n x small_n, a constant, each einsum is O(D). We
  # need to run log D einsums, thus the overall time complexity is O(D log D),
  # same as the classical divide and conquer algorithm.
  #
  # However, thanks to efficient software & hardware implementations of einsum,
  # we can often achieve far better speed than the classical algorithm on
  # accelerators, at the same time producing a far simpler XLA HLO graph.

  n = len(x)

  # Find out the shape to reshape x into.
  shape = []
  while n > 1:
    shape.append(min(n, small_n))
    n //= small_n
  shape.reverse()
  num_dims = len(shape)
  if num_dims + 1 >= 10:
    # We will run out of dimension names in einsums.
    raise ValueError(f'small_n={small_n} is too small for input size {n}')
  y = x.reshape(shape)

  # Hadamard matrices we will need.
  hadamards = dict((d, hadamard_matrix(d, x.dtype)) for d in set(shape))

  # einsum on each dimension.
  for i, d in enumerate(shape):
    y_dims = ''.join(str(j) for j in range(num_dims))
    h_dims = f'{i}{num_dims + 1}'
    out_dims = y_dims.replace(str(i), str(num_dims + 1), 1)
    operands = f'{y_dims},{h_dims}->{out_dims}'
    y = jnp.einsum(operands, y, hadamards[d], precision=precision)
  return y.flatten()


def hadamard_matrix(n: int, dtype: jnp.dtype) -> jnp.ndarray:
  """Generates the Hadamard matrix.

  Because there are JAX dtypes not supported in numpy, the equivalent function
  in scipy can't be used directly.

  Args:
    n: Number of rows/columns of the Hadamard matrix. Must be a power of 2.
    dtype: Output dtype.

  Returns:
    The Hadamard matrix of the given size and type.
  """
  return jnp.array(scipy.linalg.hadamard(n), dtype)


@jax.jit
def structured_rotation(x: jnp.ndarray,
                        rng: PRNGKey) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Computes HD(x)/sqrt(d).

  Here H is the walsh Hadamard matrix, d is the dimensionlity of x, and D
  is a random Rademacher matrix.

  Args:
    x: array to be rotated.
    rng: PRNGKey used for rotation.

  Returns:
    Rotated matrix and the original shape.
  """
  x_flat = jnp.reshape(x, [-1])
  d = 2**math.ceil(math.log2(x_flat.size))
  w = jnp.pad(x_flat, (0, d - x.size))
  rademacher = jax.random.rademacher(rng, w.shape)
  return walsh_hadamard_transform(w * rademacher) / jnp.sqrt(d), jnp.array(
      x.shape)


def inverse_structured_rotation(x: jnp.ndarray, rng: PRNGKey,
                                original_shape: jnp.ndarray) -> jnp.ndarray:
  """Computes (HD)^(-1)(x)/sqrt(d).

  Here where H is the walsh Hadamard matrix, d is the dimensionlity of x, and D
  is a random Rademacher matrix.

  Args:
    x: rotated array, which needs to be unrotated.
    rng: PRNGKey used for rotation.
    original_shape: desired shape of the output.

  Returns:
    Output of (HD)^(-1)(x)/sqrt(d) with appropriate shape.
  """
  rademacher = jax.random.rademacher(rng, x.shape)
  w = walsh_hadamard_transform(x) * rademacher / jnp.sqrt(x.size)
  original_size = jnp.prod(original_shape)
  y_flat = w.take(jnp.arange(original_size))
  return jnp.reshape(y_flat, original_shape)


def structured_rotation_pytree(params: Params,
                               rng: PRNGKey) -> Tuple[Params, Params]:
  """Applies strucuted rotation to all elements of the pytree.

  Args:
    params: pytree to be rotated.
    rng: jax random key.

  Returns:
    Pytrees of rotated arrays and shapes.
  """
  leaves, tree_def = jax.tree_util.tree_flatten(params)
  rngs = jax.random.split(rng, len(leaves))
  rotated_leaves = []
  shapes = []
  for l, r in zip(leaves, rngs):
    leaf, shape = structured_rotation(l, r)
    rotated_leaves.append(leaf)
    shapes.append(shape)
  rotated_pytree = jax.tree_util.tree_unflatten(tree_def, rotated_leaves)
  original_shapes_pytree = jax.tree_util.tree_unflatten(tree_def, shapes)
  return rotated_pytree, original_shapes_pytree


def inverse_structured_rotation_pytree(params: Params, rng: PRNGKey,
                                       shapes: Params) -> Params:
  """Applies inverse structured rotation to all elements of the pytree.

  Args:
    params: pytree to be rotated.
    rng: jax random key.
    shapes: pytree of shapes to be rotated.

  Returns:
    Inversely rotated pytree whose arrays are specified by input shapes.
  """
  leaves, tree_def = jax.tree_util.tree_flatten(params)
  leaves_shapes, _ = jax.tree_util.tree_flatten(shapes)
  rngs = jax.random.split(rng, len(leaves))
  new_leaves = []
  for l, r, shape in zip(leaves, rngs, leaves_shapes):
    new_leaves.append(inverse_structured_rotation(l, r, shape))
  return jax.tree_util.tree_unflatten(tree_def, new_leaves)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.datasets.cifar100.

This file only tests preprocessing functions.
"""

from absl.testing import absltest
from fedjax.datasets import cifar100
import numpy as np
import numpy.testing as npt


class Cifar100Test(absltest.TestCase):

  def test_preprocess_client(self):
    image = np.random.randint(256, size=[4, 32, 32, 3]).astype(np.uint8)
    coarse_label = np.random.randint(20, size=[4]).astype(np.int64)
    label = np.random.randint(100, size=[4]).astype(np.int64)
    npt.assert_equal(
        cifar100.preprocess_client(None, {
            'image': image,
            'coarse_label': coarse_label,
            'label': label
        }), {
            'x': image,
            'y': label
        })

  def test_preprocess_image(self):
    # 4 images, each with one side being 1 after processing.
    image = np.zeros([4, 32, 32, 3], dtype=np.uint8)
    # [125, 123, 114] becomes nearly 0 after processing.
    image[:, :, :, :] = [125, 123, 114]
    # [177, 174, 165] becomes nearly 1 after processing.
    image[0, 0, :, :] = [177, 174, 165]
    image[1, -1, :, :] = [177, 174, 165]
    image[2, :, 0, :] = [177, 174, 165]
    image[3, :, -1, :] = [177, 174, 165]

    with self.subTest('is_train=False'):
      processed = cifar100.preprocess_image(image, is_train=False)
      self.assertEqual(processed.dtype, np.float32)
      ref = np.zeros([4, 32, 32, 3], dtype=np.float32)
      ref[0, 0, :, :] = 1
      ref[1, -1, :, :] = 1
      ref[2, :, 0, :] = 1
      ref[3, :, -1, :] = 1
      npt.assert_allclose(processed, ref, rtol=0.01, atol=0.01)

    with self.subTest('is_train=True'):
      # Expected crop start offsets: (2, 6); flip: yes.
      np.random.seed(4)
      processed = cifar100.preprocess_image(image, is_train=True)
      self.assertEqual(processed.dtype, np.float32)
      ref = np.zeros([4, 32, 32, 3], dtype=np.float32)
      padding = [-2.4290657, -2.4182549, -2.221393]
      ref[:, :2, :, :] = padding
      ref[:, :, :2, :] = padding
      ref[0, 2, 2:, :] = 1
      ref[3, 2:, 2, :] = 1
      npt.assert_allclose(processed, ref, rtol=0.01, atol=0.01)

  def test_preprocess_batch(self):
    np.random.seed(0)
    examples = {
        'x': np.random.randint(256, size=[4, 32, 32, 3]).astype(np.uint8),
        'y': np.random.randint(100, size=[4]).astype(np.int32)
    }

    train_processed = cifar100.preprocess_batch(examples, is_train=True)
    self.assertCountEqual(train_processed, ['x', 'y'])
    self.assertIsNot(train_processed['x'], examples['x'])
    self.assertEqual(train_processed['x'].dtype, np.float32)
    self.assertEqual(train_processed['x'].shape, (4, 32, 32, 3))
    self.assertIs(train_processed['y'], examples['y'])

    eval_processed = cifar100.preprocess_batch(examples, is_train=False)
    self.assertCountEqual(eval_processed, ['x', 'y'])
    self.assertIsNot(eval_processed['x'], examples['x'])
    self.assertEqual(eval_processed['x'].dtype, np.float32)
    self.assertEqual(eval_processed['x'].shape, (4, 32, 32, 3))
    self.assertIs(eval_processed['y'], examples['y'])

    self.assertTrue(np.any(train_processed['x'] != eval_processed['x']))

  def test_preprocess_image_tff(self):
    image = np.random.RandomState(0).choice(
        256, size=(2, 32, 32, 3)).astype(np.uint8)
    crop_height = 24
    crop_width = 24

    with self.subTest('distort=False'):
      processed = cifar100.preprocess_image_tff(
          image, crop_height, crop_width, distort=False)
      self.assertEqual(processed.dtype, np.float32)
      self.assertTupleEqual(processed.shape, (2, crop_height, crop_width, 3))

    with self.subTest('distort=True'):
      processed = cifar100.preprocess_image_tff(
          image, crop_height, crop_width, distort=True)
      self.assertEqual(processed.dtype, np.float32)
      self.assertTupleEqual(processed.shape, (2, crop_height, crop_width, 3))

  def test_preprocess_image_tff_invalid_crop(self):
    image = np.zeros((2, 32, 32, 3)).astype(np.uint8)
    with self.assertRaisesRegex(
        ValueError, 'The crop_height and crop_width must be between 1 and 32'):
      cifar100.preprocess_image_tff(
          image, crop_height=33, crop_width=-1, distort=False)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated cifar100."""

import os.path
from typing import List, Optional, Tuple

from fedjax.core import client_datasets
from fedjax.core import federated_data
from fedjax.core import sqlite_federated_data
from fedjax.core import util
from fedjax.datasets import downloads

import numpy as np

tf = util.import_tf()

SPLITS = ('train', 'test')
_TFF_SQLITE_COMPRESSED_HEXDIGEST = '23d3916c9caa33395737ee560cc7cb77bbd05fc7b73647ee7be3a7e764172939'
_TFF_SQLITE_COMPRESSED_NUM_BYTES = 153019920
_FEDJAX_SQLITE_HEXDIGEST = {
    'train': 'a4dc2f4ac4c9b6e7bd5234a1b568389384c00c903dcd34001c3cf50a4a81c713',
    'test': '74416c6ee8f41b1086f0e0e3c5289ac6df5a641100df48e8b583411a55de891f'
}
_FEDJAX_SQLITE_NUM_BYTES = {'train': 140521472, 'test': 28135424}


def cite():
  """Returns BibTeX citation for the dataset."""
  return """@TECHREPORT{Krizhevsky09learningmultiple,
    author = {Alex Krizhevsky},
    title = {Learning multiple layers of features from tiny images},
    institution = {},
    year = {2009}
}"""


def _parse_tf_examples(vs: List[bytes]) -> client_datasets.Examples:
  tf_examples = tf.io.parse_example(
      vs,
      features={
          'coarse_label': tf.io.FixedLenFeature(shape=(), dtype=tf.int64),
          'image': tf.io.FixedLenFeature(shape=(32, 32, 3), dtype=tf.int64),
          'label': tf.io.FixedLenFeature(shape=(), dtype=tf.int64),
      })
  tf_examples['image'] = tf.cast(tf_examples['image'], dtype=tf.uint8)
  return tf.nest.map_structure(lambda t: t.numpy(), tf_examples)


def load_split(split: str,
               mode: str = 'sqlite',
               cache_dir: Optional[str] = None) -> federated_data.FederatedData:
  """Loads a cifar100 split.

  Features:

  - image: [N, 32, 32, 3] uint8 pixels.
  - coarse_label: [N] int64 coarse labels in the range [0, 20).
  - label: [N] int64 labels in the range [0, 100).

  Args:
    split: Name of the split. One of SPLITS.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    FederatedData.
  """
  if split not in SPLITS:
    raise ValueError(f'Invalid split={split!r}')
  if cache_dir is not None and mode != 'sqlite':
    raise ValueError('Caching locally is only supported in "sqlite" mode')
  if mode == 'sqlite':
    # Download and decompress LZMA compressed SQLite file from TFF.
    compressed_path = downloads.maybe_download(
        'https://storage.googleapis.com/tff-datasets-public/cifar100.sqlite.lzma',
        cache_dir)
    # Validate that the original TFF file has not been updated.
    downloads.validate_file(compressed_path, _TFF_SQLITE_COMPRESSED_NUM_BYTES,
                            _TFF_SQLITE_COMPRESSED_HEXDIGEST)
    decompressed_path = downloads.maybe_lzma_decompress(compressed_path)
    path = os.path.join(
        os.path.dirname(decompressed_path),
        f'federated_cifar100_{split}.sqlite')
    if os.path.exists(path):
      downloads.log(f'Reusing cached file {path!r}')
    else:
      # Convert TFF "1 example per record" to "all client examples per record".
      with sqlite_federated_data.SQLiteFederatedDataBuilder(path) as builder:
        client_ids_examples = map(
            lambda c: (c[0], c[1].all_examples()),
            sqlite_federated_data.TFFSQLiteClientsIterator(
                decompressed_path, _parse_tf_examples, split))
        builder.add_many(client_ids_examples)
      # Validate that the final produced SQLite is consistent.
      downloads.validate_file(path, _FEDJAX_SQLITE_NUM_BYTES[split],
                              _FEDJAX_SQLITE_HEXDIGEST[split])
    return sqlite_federated_data.SQLiteFederatedData.new(path)
  else:
    raise ValueError(f'Unsupported mode={mode!r}')


def load_data(
    mode: str = 'sqlite',
    cache_dir: Optional[str] = None
) -> Tuple[federated_data.FederatedData, federated_data.FederatedData]:
  """Loads partially preprocessed cifar100 splits.

  Features:

  - x: [N, 32, 32, 3] uint8 pixels.
  - y: [N] int32 labels in the range [0, 100).

  Additional preprocessing (e.g. centering and normalizing) depends on whether
  a split is used for training or eval. For example,::

    import functools
    from fedjax.datasets import cifar100
    # Load partially preprocessed splits.
    train, test = cifar100.load_data()
    # Preprocessing for training.
    train_for_train = train.preprocess_batch(
        functools.partial(preprocess_batch, is_train=True))
    # Preprocessing for eval.
    train_for_eval = train.preprocess_batch(
        functools.partial(preprocess_batch, is_train=False))
    test = test.preprocess_batch(
        functools.partial(preprocess_batch, is_train=False))

  Features after this preprocessing:

  - x: [N, 32, 32, 3] float32 preprocessed pixels.
  - y: [N] int32 labels in the range [0, 100).

  Alternatively, you can apply the same preprocessing as TensorFlow Federated
  following tff.simulation.baselines.cifar100.create_image_classification_task.
  For example,::

    from fedjax.datasets import cifar100
    train, test = cifar100.load_data()
    train = train.preprocess_batch(preprocess_batch_tff)
    test = test.preprocess_batch(preprocess_batch_tff)

  Features after this preprocessing:

  - x: [N, 24, 24, 3] float32 preprocessed pixels.
  - y: [N] int32 labels in the range [0, 100).

  Note: ``preprocess_batch`` and ``preprocess_batch_tff`` are just convenience
  wrappers around :meth:`preprocess_image` and :meth:`preprocess_image_tff`,
  respectively, for use with :meth:`fedjax.FederatedData.preprocess_batch`.

  Args:
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    A (train, test) tuple of federated data.
  """
  train = load_split('train', mode, cache_dir)
  test = load_split('test', mode, cache_dir)
  return (train.preprocess_client(preprocess_client),
          test.preprocess_client(preprocess_client))


def preprocess_client(
    client_id: federated_data.ClientId,
    examples: client_datasets.Examples) -> client_datasets.Examples:
  del client_id
  return {'x': examples['image'], 'y': examples['label'].astype(np.int32)}


# Mean and stddev computed assuming RGB values lie in [0,1].
CIFAR100_PIXELS_MEAN = np.array([0.4914, 0.4822, 0.4465], dtype=np.float32)
CIFAR100_PIXELS_INVERSE_STDDEV = (
    1 / np.array([0.2023, 0.1994, 0.2010], dtype=np.float32))


def preprocess_image(image: np.ndarray, is_train: bool) -> np.ndarray:
  """Augments and preprocesses CIFAR-100 images by cropping, flipping, and normalizing.

  Preprocessing procedure and values taken from
  `pytorch-cifar
  <https://github.com/kuangliu/pytorch-cifar/blob/49b7aa97b0c12fe0d4054e670403a16b6b834ddd/main.py#L30-L40>`_.

  Args:
    image: [N, 32, 32, 3] uint8 pixels.
    is_train: Whether we are preprocessing for training or eval.

  Returns:
    Processed [N, 32, 32, 3] float32 pixels.
  """
  if is_train:
    # Pad 4 zero pixels on all sides and then randomly crop back to (32, 32, 3).
    num_paddings = 4
    image = np.pad(
        image, [(0, 0), (num_paddings, num_paddings),
                (num_paddings, num_paddings), (0, 0)],
        mode='constant')
    # Start offsets for cropping.
    i, j = np.random.randint(num_paddings * 2, size=[2])
    image = image[:, i:i + 32, j:j + 32, :]
    # Random horizontal flip.
    if np.random.randint(2):
      image = np.flip(image, axis=-2)
  # Center and normalize.
  image = ((image.astype(np.float32) / 255 - CIFAR100_PIXELS_MEAN) *
           CIFAR100_PIXELS_INVERSE_STDDEV)
  return image


def preprocess_batch(examples: client_datasets.Examples,
                     is_train: bool) -> client_datasets.Examples:
  return {'x': preprocess_image(examples['x'], is_train), 'y': examples['y']}


def preprocess_image_tff(image: np.ndarray, crop_height: int, crop_width: int,
                         distort: bool) -> np.ndarray:
  """Augments and preprocesses CIFAR-100 images following TensorFlow Federated."""
  if crop_height < 1 or crop_width < 1 or crop_height > 32 or crop_width > 32:
    raise ValueError('The crop_height and crop_width must be between 1 and 32.')

  if distort:
    # Distort via random crops and flips.
    # Start offsets for cropping should match tf.image.random_crop.
    crop_shape = np.array((crop_height, crop_width, 3)).astype(np.int32)
    shape = np.array(image.shape).astype(np.int32)[1:]
    limit = shape - crop_shape + 1
    offset = np.random.uniform(
        high=np.iinfo(limit.dtype).max, size=shape.shape).astype(
            limit.dtype) % limit
    begin_i, begin_j, _ = offset
    end_i, end_j, _ = offset + crop_shape
    image = image[:, begin_i:end_i, begin_j:end_j, :]
    # Random horizontal flip should match tf.image.random_flip_left_right.
    if np.random.randint(2):
      image = np.flip(image, axis=-2)
  else:
    # Center crop should match tf.image.resize_with_crop_or_pad.
    height_offset = (32 - crop_height) // 2
    width_offset = (32 - crop_width) // 2
    image = image[:, height_offset:height_offset + crop_height,
                  width_offset:width_offset + crop_width, :]
  image = image.astype(np.float32)
  # Normalize should match tf.image.per_image_standardization.
  num_pixels = np.prod(image.shape[-3:], dtype=np.float32)
  image_mean = np.mean(image, axis=(-1, -2, -3), keepdims=True)
  image_std = np.std(image, axis=(-1, -2, -3), keepdims=True)
  image_adjusted_std = np.maximum(image_std, np.sqrt(num_pixels))
  return (image - image_mean) / image_adjusted_std


def preprocess_batch_tff(examples: client_datasets.Examples,
                         crop_height: int = 24,
                         crop_width: int = 24,
                         distort: bool = False) -> client_datasets.Examples:
  """Preprocesses CIFAR-100 examples following TensorFlow Federated.

  Preprocessing procedure and values taken from
  `tff.simulation.baselines.cifar100.create_image_classification_task
  <https://www.tensorflow.org/federated/api_docs/python/tff/simulation/baselines/cifar100/create_image_classification_task>`_.

  Args:
    examples: Batch of examples containing
      - x: [N, 32, 32, 3] uint8 pixels.
      - y: [N] int32 labels in the range [0, 100).
    crop_height: Desired height for cropping images. Must be between 1 and 32.
      Defaults to the value used in TensorFlow Federated.
    crop_width: Desired width for cropping images. Must be between 1 and 32.
      Defaults to the value used in TensorFlow Federated.
    distort: Whether to distort the image via random crops and flips. Defaults
      to center cropping.

  Returns:
    Processed batch of examples containing
      - x: [N, ``crop_height``, ``crop_width``, 3] float32 pixels.
      - y: [N] int32 labels in the range [0, 100).
  """
  return {
      'x':
          preprocess_image_tff(examples['x'], crop_height, crop_width, distort),
      'y':
          examples['y'],
  }



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.algorithms.fed_prox."""
from absl.testing import absltest

from fedjax.algorithms import fed_prox
from fedjax.core import client_datasets
from fedjax.core import optimizers
from fedjax.core import tree_util

import jax
import jax.numpy as jnp
import numpy.testing as npt


def per_example_loss(params, batch, rng):
  del rng
  print(batch['x'])
  print(params['w'])
  return batch['x'] * params['w']


class FedProxTest(absltest.TestCase):

  def test_fed_prox(self):
    client_optimizer = optimizers.sgd(learning_rate=1.0)
    server_optimizer = optimizers.sgd(learning_rate=1.0)
    client_batch_hparams = client_datasets.ShuffleRepeatBatchHParams(
        batch_size=2, num_epochs=1, seed=0)
    algorithm = fed_prox.fed_prox(
        per_example_loss,
        client_optimizer,
        server_optimizer,
        client_batch_hparams,
        proximal_weight=0.01)

    with self.subTest('init'):
      state = algorithm.init({'w': jnp.array(4.)})
      npt.assert_array_equal(state.params['w'], 4.)
      self.assertLen(state.opt_state, 2)

    with self.subTest('apply'):
      clients = [
          (b'cid0',
           client_datasets.ClientDataset({'x': jnp.array([2., 4., 6.])}),
           jax.random.PRNGKey(0)),
          (b'cid1',
           client_datasets.ClientDataset({'x': jnp.array([8., 10.])}),
           jax.random.PRNGKey(1)),
      ]
      state, client_diagnostics = algorithm.apply(state, clients)
      npt.assert_allclose(state.params['w'], -3.77)
      npt.assert_allclose(client_diagnostics[b'cid0']['delta_l2_norm'], 6.95)
      npt.assert_allclose(client_diagnostics[b'cid1']['delta_l2_norm'], 9.)

  def test_create_train_for_each_client(self):
    proximal_weight = 0.01

    def fed_prox_loss(params, server_params, batch, rng):
      example_loss = per_example_loss(params, batch, rng)
      proximal_loss = 0.5 * proximal_weight * tree_util.tree_l2_squared(
          jax.tree_util.tree_multimap(lambda a, b: a - b, server_params,
                                      params))
      return jnp.mean(example_loss + proximal_loss)

    grad_fn = jax.grad(fed_prox_loss)
    client_optimizer = optimizers.sgd(learning_rate=1.0)
    train_for_each_client = fed_prox.create_train_for_each_client(
        grad_fn, client_optimizer)
    batched_clients = [
        (b'cid0',
         [{'x': jnp.array([2., 4., 6.])}, {'x': jnp.array([8., 10., 12.])}],
         jax.random.PRNGKey(0)),
        (b'cid1',
         [{'x': jnp.array([1., 3., 5.])}, {'x': jnp.array([7., 9., 11.])}],
         jax.random.PRNGKey(1)),
    ]
    server_params = {'w': jnp.array(4.0)}
    client_outputs = dict(train_for_each_client(server_params, batched_clients))
    npt.assert_allclose(client_outputs[b'cid0']['w'], 13.96)
    npt.assert_allclose(client_outputs[b'cid1']['w'], 11.97)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EMNIST models."""

from fedjax.core import metrics
from fedjax.core import models

import haiku as hk
import jax
try:
  from jax.example_libraries import stax
except ModuleNotFoundError:
  from jax.experimental import stax
import jax.numpy as jnp
import numpy as np


class Dropout(hk.Module):
  """Dropout haiku module."""

  def __init__(self, rate: float = 0.5):
    """Initializes dropout module.

    Args:
      rate: Probability that each element of x is discarded. Must be in [0, 1).
    """
    super().__init__()
    self._rate = rate

  def __call__(self, x: jnp.ndarray, is_train: bool):
    if is_train:
      return hk.dropout(rng=hk.next_rng_key(), rate=self._rate, x=x)
    return x


class ConvDropoutModule(hk.Module):
  """Custom haiku module for CNN with dropout.

  This must be defined as a custom hk.Module because only a single positional
  argument is allowed when using hk.Sequential.
  """

  def __init__(self, num_classes):
    super().__init__()
    self._num_classes = num_classes

  def __call__(self, x: jnp.ndarray, is_train: bool):
    x = hk.Conv2D(output_channels=32, kernel_shape=(3, 3), padding='VALID')(x)
    x = jax.nn.relu(x)
    x = hk.Conv2D(output_channels=64, kernel_shape=(3, 3), padding='VALID')(x)
    x = jax.nn.relu(x)
    x = (
        hk.MaxPool(
            window_shape=(1, 2, 2, 1), strides=(1, 2, 2, 1),
            padding='VALID')(x))
    x = Dropout(rate=0.25)(x, is_train)
    x = hk.Flatten()(x)
    x = hk.Linear(128)(x)
    x = jax.nn.relu(x)
    x = Dropout(rate=0.5)(x, is_train)
    x = hk.Linear(self._num_classes)(x)
    return x


# Defines the expected structure of input batches to the model. This is used to
# determine the model parameter shapes.
_HAIKU_SAMPLE_BATCH = {
    'x': np.zeros((1, 28, 28, 1), dtype=np.float32),
    'y': np.zeros(1, dtype=np.float32)
}
_STAX_SAMPLE_SHAPE = (-1, 28, 28, 1)

_TRAIN_LOSS = lambda b, p: metrics.unreduced_cross_entropy_loss(b['y'], p)
_EVAL_METRICS = {
    'loss': metrics.CrossEntropyLoss(),
    'accuracy': metrics.Accuracy()
}


def create_conv_model(only_digits: bool = False) -> models.Model:
  """Creates EMNIST CNN model with dropout with haiku.

  Matches the model used in:

  Adaptive Federated Optimization
    Sashank Reddi, Zachary Charles, Manzil Zaheer, Zachary Garrett, Keith Rush,
    Jakub Konečný, Sanjiv Kumar, H. Brendan McMahan.
    https://arxiv.org/abs/2003.00295

  Args:
    only_digits: Whether to use only digit classes [0-9] or include lower and
      upper case characters for a total of 62 classes.

  Returns:
    Model
  """
  num_classes = 10 if only_digits else 62

  def forward_pass(batch, is_train=True):
    return ConvDropoutModule(num_classes)(batch['x'], is_train)

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS,
      # is_train determines whether to apply dropout or not.
      train_kwargs={'is_train': True},
      eval_kwargs={'is_train': False})


def create_dense_model(only_digits: bool = False,
                       hidden_units: int = 200) -> models.Model:
  """Creates EMNIST dense net with haiku."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(hidden_units),
        jax.nn.relu,
        hk.Linear(hidden_units),
        jax.nn.relu,
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)


def create_logistic_model(only_digits: bool = False) -> models.Model:
  """Creates EMNIST logistic model with haiku."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return models.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_HAIKU_SAMPLE_BATCH,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)


def create_stax_dense_model(only_digits: bool = False,
                            hidden_units: int = 200) -> models.Model:
  """Creates EMNIST dense net with stax."""
  num_classes = 10 if only_digits else 62
  stax_init, stax_apply = stax.serial(stax.Flatten,
                                      stax.Dense(hidden_units), stax.Relu,
                                      stax.Dense(hidden_units), stax.Relu,
                                      stax.Dense(num_classes))
  return models.create_model_from_stax(
      stax_init=stax_init,
      stax_apply=stax_apply,
      sample_shape=_STAX_SAMPLE_SHAPE,
      train_loss=_TRAIN_LOSS,
      eval_metrics=_EVAL_METRICS)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple download manager."""

import hashlib
import math
import os.path
import shutil
import sys
import time
from typing import Callable, Iterator, Optional
import urllib.parse

import lzma
import requests


def progress(n: int) -> Iterator[int]:
  """A simple generator for tracking progress."""
  start = time.time()
  last_log = -1
  for i in range(n):
    yield i
    elapsed = time.time() - start
    if elapsed - last_log >= 1:
      last_log = elapsed
      eta = elapsed / (i + 1) * (n - i - 1)
      log(f'\r{100*(i + 1)/n:3.0f}%, ETA: {format_duration(eta)}', end='')
  log(f'\r100%, elapsed: {format_duration(time.time() - start)}')


def maybe_download(url: str,
                   cache_dir: Optional[str] = None,
                   progress_: Callable[[int], Iterator[int]] = progress) -> str:
  """Downloads `url` to local disk.

  Args:
    url: URL to download from.
    cache_dir: Where to cache the file. If None, uses default_cache_dir().
    progress_: A callable that yields like range(n), for tracking progress.

  Returns:
    Path to local file.
  """
  # TODO(wuke): Avoid race conditions when downloading the same file from
  # different threads/processes at the same time.
  if cache_dir is None:
    cache_dir = default_cache_dir()
  os.makedirs(cache_dir, exist_ok=True)
  path = os.path.join(cache_dir,
                      os.path.basename(urllib.parse.urlparse(url).path))
  if os.path.exists(path):
    log(f'Reusing cached file {path!r}')
  else:
    log(f'Downloading {url!r} to {path!r}')
    with open(path + '.partial', 'wb') as fo:
      r = requests.get(url, stream=True)
      r.raise_for_status()
      length = int(r.headers['content-length'])
      block_size = 1 << 18
      for _ in progress_((length + block_size - 1) // block_size):
        fo.write(r.raw.read(block_size))
    os.rename(path + '.partial', path)
  return path


def format_duration(seconds: float) -> str:
  """Formats duration in seconds into hours/minutes/seconds."""
  if seconds < 60:
    return f'{seconds:.0f}s'
  elif seconds < 3600:
    minutes = math.floor(seconds / 60)
    seconds -= minutes * 60
    return f'{minutes}m{seconds:.0f}s'
  else:
    hours = math.floor(seconds / 3600)
    seconds -= hours * 3600
    minutes = math.floor(seconds / 60)
    seconds -= minutes * 60
    return f'{hours}h{minutes}m{seconds:.0f}s'


def log(*args, **kwargs):
  print(*args, flush=True, file=sys.stderr, **kwargs)


def default_cache_dir() -> str:
  """Returns default local file cache directory."""
  running_on_colab = 'google.colab' in sys.modules
  if running_on_colab:
    base_dir = '/tmp'
  else:
    base_dir = os.path.expanduser('~')
  cache_dir = os.path.join(base_dir, '.cache/fedjax')
  return cache_dir


def maybe_lzma_decompress(path) -> str:
  """Decompresses LZMA compressed local file or reuses cached file."""
  decompressed_path, ext = os.path.splitext(path)
  if ext != '.lzma':
    raise ValueError(
        'Only decompressing LZMA files is supported. If the file '
        'is LZMA compressed, rename the url to have a .lzma suffix.')
  if os.path.exists(decompressed_path):
    log(f'Reusing cached file {decompressed_path!r}')
  else:
    log(f'Decompressing {path!r} to {decompressed_path!r}')
    with lzma.open(path, 'rb') as fi:
      with open(decompressed_path, 'wb') as fo:
        shutil.copyfileobj(fi, fo)
  return decompressed_path


def validate_file(path: str, expected_num_bytes: int, expected_hexdigest: str):
  """Validates path file contents has specified number of bytes and hash."""
  with open(path, 'rb') as f:
    data = f.read()
  num_bytes = len(data)
  if num_bytes != expected_num_bytes:
    raise ValueError(
        f'Expected file content number of bytes to be {expected_num_bytes} but found {num_bytes}.'
    )
  hexdigest = hashlib.sha256(data).hexdigest()
  if hexdigest != expected_hexdigest:
    raise ValueError(
        f'Expected file content hash to be {expected_hexdigest!r} but found {hexdigest!r}.'
    )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for walsh_hadamard."""

from absl.testing import absltest
from fedjax.aggregators import walsh_hadamard
import jax
import jax.numpy as jnp
import numpy.testing as npt


class WalshHadamardTest(absltest.TestCase):

  def test_walsh_hadamard_transform(self):
    # TODO(wuke): Change bfloat16_3x to high once the PyPI JAX release catches
    # up.
    for precision in ['fastest', 'bfloat16_3x', 'highest']:
      for seed in range(5):
        n = 2**10
        x = jnp.array(jax.random.normal(jax.random.PRNGKey(seed), shape=[n]))
        expect = jnp.dot(
            walsh_hadamard.hadamard_matrix(n, x.dtype), x, precision=precision)
        for m in [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]:
          small_n = 2**m
          with self.subTest(
              f'precision={precision}, seed={seed}, small_n={small_n}'):
            y = walsh_hadamard.walsh_hadamard_transform(x, small_n)
            self.assertEqual(x.shape, y.shape)
            self.assertEqual(x.dtype, y.dtype)
            npt.assert_allclose(y, expect, rtol=1e-4, atol=1e-4)

  def test_structured_rotation(self):
    n = 10
    m = 5
    x = jnp.array(jax.random.normal(jax.random.PRNGKey(100), shape=[m, n]))
    rng = jax.random.PRNGKey(10)
    y, x_shape = walsh_hadamard.structured_rotation(x, rng)
    self.assertEqual(x.shape, x_shape)
    z = walsh_hadamard.inverse_structured_rotation(y, rng, x_shape)
    npt.assert_allclose(z, x, rtol=1e-4, atol=1e-4)

  def test_structured_rotation_pytree(self):
    params = {
        'a': jnp.array([[1.0, 0.0, 0.0], [1.0, 2.0, 3.0]]),
        'b': jnp.array([[1.0, 0.0], [1.0, 2.0]])
    }
    rng = jax.random.PRNGKey(10)
    y, shapes = walsh_hadamard.structured_rotation_pytree(params, rng)
    z = walsh_hadamard.inverse_structured_rotation_pytree(y, rng, shapes)
    npt.assert_allclose(z['a'], params['a'], rtol=1e-4, atol=1e-4)
    npt.assert_allclose(z['b'], params['b'], rtol=1e-4, atol=1e-4)


  if __name__ == '__main__':
    absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.Let's use the full copyright notice
"""Test for data_to_sqlite dialogue normalization."""

from absl.testing import absltest
from absl.testing import parameterized
from fedjax.datasets.scripts.cornell_movie_dialogs import data_to_sqlite


class TxtToSQLiteTest(parameterized.TestCase):
  @parameterized.named_parameters(
      {
          'testcase_name': 'test hyphenated',
          'input_string': 'trigger-happy?',
          'desired': 'trigger-happy',
      }, {
          'testcase_name': 'test sentence',
          'input_string': 'Thank God! If I had to hear one more story...\n',
          'desired': 'thank god if i had to hear one more story',
      }, {
          'testcase_name': 'test decimal',
          'input_string': 'Pi is 3.14 and pi day is in 10      months.',
          'desired': 'pi is <NUMERIC> and pi day is in <NUMERIC> months',
      }, {
          'testcase_name': 'test dollar',
          'input_string': 'I have 50,000.00 dollars in my account right now.\n',
          'desired': 'i have <NUMERIC> dollars in my account right now',
      }, {
          'testcase_name': 'test hyphen',
          'input_string': 'This is a double--hyphen! And this is a single-one.',
          'desired': 'this is a double hyphen and this is a single-one',
      }, {
          'testcase_name': 'test ellipses',
          'input_string': 'I..do...not......know.....\n',
          'desired': 'i do not know',
      }, {
          'testcase_name':
              'test html tags',
          'input_string':
              '<u>These are</u> underlining HTML and <b>bold</b> HTML tags.',
          'desired':
              'these are underlining html and bold html tags',
      })

  def test_normalize_dialogue_string(self, input_string, desired):
    self.assertEqual(
        desired,
        data_to_sqlite.normalize_dialogue_string(input_string))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Algorithms for model update compression."""

import itertools
import math
from typing import Iterable, Optional, Tuple

from fedjax.aggregators import aggregator
from fedjax.aggregators import walsh_hadamard
from fedjax.core import dataclasses
from fedjax.core import tree_util
from fedjax.core.federated_data import ClientId
from fedjax.core.typing import PRNGKey, Params

import haiku as hk
import jax
import jax.numpy as jnp


@dataclasses.dataclass
class CompressionState:
  """State of compression aggregator passed between rounds.

  Attributes:
    num_bits: number of bits transmitted per client.
    rng: random key used for compression.
  """
  num_bits: float
  rng: PRNGKey


def binary_stochastic_quantize(v: jnp.ndarray,
                               rng: PRNGKey,
                               v_min: Optional[float] = None,
                               v_max: Optional[float] = None) -> jnp.ndarray:
  """Binary stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    v: vector to be quantized.
    rng: jax random key.
    v_min: minimum threshold for quantization. If None, sets it to jnp.amin(v).
    v_max: maximum threshold for quantization. If None, sets it to jnp.amax(v).

  Returns:
    Quantized array.
  """
  if v_min is None:
    v_min = jnp.amin(v)
  if v_max is None:
    v_max = jnp.amax(v)
  v = jnp.nan_to_num((v - v_min) / (v_max - v_min))
  v = jnp.maximum(0., jnp.minimum(v, 1.))
  rand = jax.random.uniform(key=rng, shape=v.shape)
  return jnp.where(rand > v, v_min, v_max)


def uniform_stochastic_quantize(v: jnp.ndarray,
                                num_levels: int,
                                rng: PRNGKey,
                                v_min: Optional[float] = None,
                                v_max: Optional[float] = None) -> jnp.ndarray:
  """Uniform stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    v: vector to be quantized.
    num_levels: Number of levels of quantization.
    rng: jax random key.
    v_min: minimum threshold for quantization. If None, sets it to jnp.amin(v).
    v_max: maximum threshold for quantization. If None, sets it to jnp.amax(v).

  Returns:
    Quantized array.
  """
  # Rescale the vector to be between zero to one.
  if v_min is None:
    v_min = jnp.amin(v)
  if v_max is None:
    v_max = jnp.amax(v)
  v = jnp.nan_to_num((v - v_min) / (v_max - v_min))
  v = jnp.maximum(0., jnp.minimum(v, 1.))
  # Compute the upper and lower boundary of each value.
  v_ceil = jnp.ceil(v * (num_levels - 1)) / (num_levels - 1)
  v_floor = jnp.floor(v * (num_levels - 1)) / (num_levels - 1)
  # uniformly quantize between v_ceil and v_floor.
  rand = jax.random.uniform(key=rng, shape=v.shape)
  threshold = jnp.nan_to_num((v - v_floor) / (v_ceil - v_floor))
  quantized = jnp.where(rand > threshold, v_floor, v_ceil)
  # Rescale the values and return it.
  return v_min + quantized * (v_max - v_min)


@jax.jit
def uniform_stochastic_quantize_pytree(params: Params, num_levels: int,
                                       rng: PRNGKey) -> Params:
  """Applies uniform_stochastic_quantize for all leafs of the pytree.

  Args:
    params: pytree to be quantized.
    num_levels: Number of levels of quantization.
    rng: jax random key.

  Returns:
    Quantized pytree.
  """
  leaves, tree_def = jax.tree_util.tree_flatten(params)
  rngs = jax.random.split(rng, len(leaves))
  new_leaves = []
  for l, r in zip(leaves, rngs):
    new_leaves.append(uniform_stochastic_quantize(l, num_levels, r))
  return jax.tree_util.tree_unflatten(tree_def, new_leaves)


def num_leaves(pytree):
  return len(jax.tree_util.tree_leaves(pytree))


@jax.jit
def _entropy(v, uniq):
  uniq = jnp.concatenate([uniq, jnp.array([jnp.inf])], axis=0)
  hist, _ = jnp.histogram(v, bins=uniq)
  hist = hist / jnp.sum(hist)
  entropy = -jnp.sum(hist * jnp.log2(hist))
  return entropy


@jax.jit
def _hist_bits(v, uniq):
  """Number of bits required to encode the histogram of v."""
  d = v.size
  k = uniq.size
  return k * jnp.log2(jnp.exp(1)*(d + k)/k)


def arithmetic_encoding_num_bits(v: jnp.ndarray) -> int:
  """Computes number of bits needed to store v via arithmetic coding."""
  v = jnp.nan_to_num(v)
  v = v.flatten()
  uniq = jnp.unique(v)
  entropy = _entropy(v, uniq)
  hist_bits = _hist_bits(v, uniq)
  return hist_bits + (v.size * entropy) + (2 * 32) + 2


def uniform_stochastic_quantizer(
    num_levels: int,
    rng: PRNGKey,
    encode_algorithm: Optional[str] = None) -> aggregator.Aggregator:
  """Returns (weighted) mean of input uniformly quantized trees using the

  uniform stochastic algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    num_levels: number of levels of quantization.
    rng: PRNGKey used for compression:
    encode_algorithm: None or arithmetic

  Returns:
    Compression aggregator.
  """

  def init():
    return CompressionState(0.0, rng)

  def apply(
      clients_params_and_weights: Iterable[Tuple[ClientId, Params, float]],
      aggregator_state: CompressionState) -> Tuple[Params, CompressionState]:

    if encode_algorithm is not None:
      assert encode_algorithm == 'arithmetic'

    def quantize_params_and_weight(client_params_and_weight, rng):
      _, params, weight = client_params_and_weight
      return uniform_stochastic_quantize_pytree(params, num_levels, rng), weight

    rng, use_rng = jax.random.split(aggregator_state.rng)
    # TODO(theertha): remove the usage of hk.PRNGSequence.
    rng_seq = hk.PRNGSequence(use_rng)
    clients_params_and_weight_rng = zip(clients_params_and_weights, rng_seq)
    quantized_p_and_w = itertools.starmap(quantize_params_and_weight,
                                          clients_params_and_weight_rng)
    new_bits = 0.
    if encode_algorithm == 'arithmetic':
      # Accumulate the number of bits used by all clients without loading the
      # entire iterator into memory at once.
      total_bits = []

      def arithmetic_encoding_num_bits_pytree(params, weights):
        leaves, _ = jax.tree_util.tree_flatten(params)
        bits = sum([arithmetic_encoding_num_bits(leaf) for leaf in leaves])
        total_bits.append(bits)
        return params, weights

      quantized_p_and_w = itertools.starmap(arithmetic_encoding_num_bits_pytree,
                                            quantized_p_and_w)
      aggregated_params = tree_util.tree_mean(quantized_p_and_w)
      new_bits = sum(total_bits) / len(total_bits) if len(total_bits) else 0.

    else:
      aggregated_params = tree_util.tree_mean(quantized_p_and_w)
      total_num_params = tree_util.tree_size(aggregated_params)
      total_num_floats = 2 * num_leaves(aggregated_params)
      # 32 bits for every float and log2(num_levels) bit for every parameter.
      new_bits = math.log2(
          num_levels) * total_num_params + 32 * total_num_floats
    new_state = CompressionState(aggregator_state.num_bits + new_bits, rng)
    return aggregated_params, new_state

  return aggregator.Aggregator(init, apply)


def rotated_uniform_stochastic_quantizer(num_levels: int,
                                         rng: PRNGKey) -> aggregator.Aggregator:
  """Returns (weighted) mean of input uniformly quantized trees with rotation

  using the algorithm in https://arxiv.org/pdf/1611.00429.pdf.

  Args:
    num_levels: number of levels of quantization.
    rng: PRNGKey used for compression.

  Returns:
    Compression aggregator.
  """

  def init():
    return CompressionState(0.0, rng)

  def apply(
      clients_params_and_weights: Iterable[Tuple[ClientId, Params, float]],
      aggregator_state: CompressionState) -> Tuple[Params, CompressionState]:

    rng, rotation_rng = jax.random.split(aggregator_state.rng)

    def quantize_params_and_weight(client_params_and_weight, rng):
      _, params, weight = client_params_and_weight
      params, shapes = walsh_hadamard.structured_rotation_pytree(
          params, rotation_rng)
      params = walsh_hadamard.inverse_structured_rotation_pytree(
          uniform_stochastic_quantize_pytree(params, num_levels, rng),
          rotation_rng, shapes)
      return params, weight

    rng, use_rng = jax.random.split(rng)
    # TODO(theertha): remove the usage of hk.PRNGSequence.
    rng_seq = hk.PRNGSequence(use_rng)
    clients_params_and_weight_rng = zip(clients_params_and_weights, rng_seq)
    quantized_p_and_w = itertools.starmap(quantize_params_and_weight,
                                          clients_params_and_weight_rng)
    aggregated_params = tree_util.tree_mean(quantized_p_and_w)
    total_num_params = tree_util.tree_size(aggregated_params)
    total_num_floats = 2 * num_leaves(aggregated_params)
    # 32 bits for every float used and log2(num_levels) bit for every parameter.
    new_bits = math.log2(num_levels) * total_num_params + 32 * total_num_floats
    new_state = CompressionState(aggregator_state.num_bits + new_bits, rng)
    return aggregated_params, new_state

  return aggregator.Aggregator(init, apply)


@jax.jit
def drive_pytree(params: Params) -> Params:
  """Runs DRIVE quantization on a given pytree."""
  leaves, tree_def = jax.tree_util.tree_flatten(params)
  new_leaves = []
  for leaf in leaves:
    new_leaves.append(jnp.sum(jnp.abs(leaf)) * jnp.sign(leaf) / leaf.size)
  return jax.tree_util.tree_unflatten(tree_def, new_leaves)


def structured_drive_quantizer(rng: PRNGKey) -> aggregator.Aggregator:
  """Returns (weighted) mean using the structured DRIVE algorithm.

  Paper: https://arxiv.org/pdf/2105.08339.pdf.

  Args:
    rng: PRNGKey used for compression.

  Returns:
    Compression aggregator.
  """

  def init():
    return CompressionState(0.0, rng)

  def apply(
      clients_params_and_weights: Iterable[Tuple[ClientId, Params, float]],
      aggregator_state: CompressionState) -> Tuple[Params, CompressionState]:

    rng, rotation_rng = jax.random.split(aggregator_state.rng)

    def quantize_params_and_weight(client_id, params, weight):
      del client_id
      rotated_param, shapes = walsh_hadamard.structured_rotation_pytree(
          params, rotation_rng)
      return walsh_hadamard.inverse_structured_rotation_pytree(
          drive_pytree(rotated_param), rotation_rng, shapes), weight

    quantized_p_and_w = itertools.starmap(quantize_params_and_weight,
                                          clients_params_and_weights)

    aggregated_params = tree_util.tree_mean(quantized_p_and_w)

    total_num_params = tree_util.tree_size(aggregated_params)
    total_num_floats = 2 * num_leaves(aggregated_params)
    # 32 bits for every float used and one bit for every parameter.
    new_bits = total_num_params + 32 * total_num_floats
    new_state = CompressionState(aggregator_state.num_bits + new_bits, rng)
    return aggregated_params, new_state

  return aggregator.Aggregator(init, apply)


def terngrad_quantize(v: jnp.ndarray, rng: PRNGKey) -> jnp.ndarray:
  """Terngrad algorithm https://arxiv.org/abs/1705.07878.

  Args:
    v: vector to be quantized.
    rng: jax random key.

  Returns:
    Quantized array.
  """
  sigma = jnp.std(v)
  v = jnp.where(jnp.abs(v) > 2.5 * sigma, 2.5 * sigma * jnp.sign(v), v)
  return binary_stochastic_quantize(jnp.abs(v), rng, 0., jnp.amax(
      jnp.abs(v))) * jnp.sign(v)


@jax.jit
def terngrad_quantize_pytree(params: Params, rng: PRNGKey) -> Params:
  """Applies terngrad_quantize for all leafs of the pytree.

  Args:
    params: pytree to be quantized.
    rng: jax random key.

  Returns:
    Quantized pytree.
  """
  leaves, tree_def = jax.tree_util.tree_flatten(params)
  rngs = jax.random.split(rng, len(leaves))
  new_leaves = []
  for l, r in zip(leaves, rngs):
    new_leaves.append(terngrad_quantize(l, r))
  return jax.tree_util.tree_unflatten(tree_def, new_leaves)


def terngrad_quantizer(rng: PRNGKey) -> aggregator.Aggregator:
  """Returns (weighted) mean of terngrad algorithm.

  Paper: https://arxiv.org/abs/1705.07878.

  Args:
    rng: PRNGKey used for compression.

  Returns:
    Compression aggregator.
  """

  def init():
    return CompressionState(0.0, rng)

  def apply(
      clients_params_and_weights: Iterable[Tuple[ClientId, Params, float]],
      aggregator_state: CompressionState) -> Tuple[Params, CompressionState]:

    def quantize_params_and_weight(client_params_and_weight, rng):
      _, params, weight = client_params_and_weight
      return terngrad_quantize_pytree(params, rng), weight

    rng, use_rng = jax.random.split(aggregator_state.rng)
    # TODO(theertha): remove the usage of hk.PRNGSequence.
    rng_seq = hk.PRNGSequence(use_rng)
    clients_params_and_weight_rng = zip(clients_params_and_weights, rng_seq)
    quantized_p_and_w = itertools.starmap(quantize_params_and_weight,
                                          clients_params_and_weight_rng)
    aggregated_params = tree_util.tree_mean(quantized_p_and_w)
    total_num_params = tree_util.tree_size(aggregated_params)
    total_num_floats = 2 * num_leaves(aggregated_params)
    # 32 bits for every float used and log2(3) bit for every parameter.
    new_bits = math.log2(3) * total_num_params + 32 * total_num_floats
    new_state = CompressionState(aggregator_state.num_bits + new_bits, rng)
    return aggregated_params, new_state

  return aggregator.Aggregator(init, apply)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FedProx implementation using fedjax.core.

Federated Optimization in Heterogeneous Networks
    Tian Li, Anit Kumar Sahu, Manzil Zaheer, Maziar Sanjabi, Ameet Talwalkar,
    Virginia Smith. MLSys 2020.
    https://arxiv.org/abs/1812.06127
"""

from typing import Any, Callable, Mapping, Sequence, Tuple

from fedjax.algorithms import fed_avg

from fedjax.core import client_datasets
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import optimizers
from fedjax.core import tree_util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey

import jax
import jax.numpy as jnp

Grads = Params
ServerState = fed_avg.ServerState


def create_train_for_each_client(grad_fn, client_optimizer):
  """Builds client_init, client_step, client_final for for_each_client."""

  def client_init(server_params, client_rng):
    opt_state = client_optimizer.init(server_params)
    client_step_state = {
        'params': server_params,
        'opt_state': opt_state,
        'rng': client_rng,
        'server_params': server_params,
    }
    return client_step_state

  def client_step(client_step_state, batch):
    rng, use_rng = jax.random.split(client_step_state['rng'])
    grads = grad_fn(client_step_state['params'],
                    client_step_state['server_params'], batch, use_rng)
    opt_state, params = client_optimizer.apply(grads,
                                               client_step_state['opt_state'],
                                               client_step_state['params'])
    next_client_step_state = {
        'params': params,
        'opt_state': opt_state,
        'rng': rng,
        'server_params': client_step_state['server_params'],
    }
    return next_client_step_state

  def client_final(server_params, client_step_state):
    delta_params = jax.tree_util.tree_multimap(lambda a, b: a - b,
                                               server_params,
                                               client_step_state['params'])
    return delta_params

  return for_each_client.for_each_client(client_init, client_step, client_final)


def fed_prox(per_example_loss: Callable[[Params, BatchExample, PRNGKey],
                                        jnp.ndarray],
             client_optimizer: optimizers.Optimizer,
             server_optimizer: optimizers.Optimizer,
             client_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
             proximal_weight: float) -> federated_algorithm.FederatedAlgorithm:
  """Builds FedProx.

  Args:
    per_example_loss: A function from (params, batch, rng) to a vector of per
      example loss values. This will be combined with a proximal term based on
      server params weighted by proximal_weight.
    client_optimizer: Optimizer for local client training.
    server_optimizer: Optimizer for server update.
    client_batch_hparams: Hyperparameters for batching client dataset for train.
    proximal_weight: Weight for proximal term. 0 weight is FedAvg.

  Returns:
    FederatedAlgorithm
  """

  def fed_prox_loss(params, server_params, batch, rng):
    example_loss = per_example_loss(params, batch, rng)
    proximal_loss = 0.5 * proximal_weight * tree_util.tree_l2_squared(
        jax.tree_util.tree_multimap(lambda a, b: a - b, server_params, params))
    return jnp.mean(example_loss + proximal_loss)

  grad_fn = jax.grad(fed_prox_loss)
  train_for_each_client = create_train_for_each_client(grad_fn,
                                                       client_optimizer)

  def init(params: Params) -> ServerState:
    opt_state = server_optimizer.init(params)
    return ServerState(params, opt_state)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]]
  ) -> Tuple[ServerState, Mapping[federated_data.ClientId, Any]]:
    client_num_examples = {cid: len(cds) for cid, cds, _ in clients}
    batch_clients = [(cid, cds.shuffle_repeat_batch(client_batch_hparams), crng)
                     for cid, cds, crng in clients]
    client_diagnostics = {}
    # Running weighted mean of client updates. We do this iteratively to avoid
    # loading all the client outputs into memory since they can be prohibitively
    # large depending on the model parameters size.
    delta_params_sum = tree_util.tree_zeros_like(server_state.params)
    num_examples_sum = 0.
    for client_id, delta_params in train_for_each_client(
        server_state.params, batch_clients):
      num_examples = client_num_examples[client_id]
      delta_params_sum = tree_util.tree_add(
          delta_params_sum, tree_util.tree_weight(delta_params, num_examples))
      num_examples_sum += num_examples
      # We record the l2 norm of client updates as an example, but it is not
      # required for the algorithm.
      client_diagnostics[client_id] = {
          'delta_l2_norm': tree_util.tree_l2_norm(delta_params)
      }
    mean_delta_params = tree_util.tree_inverse_weight(delta_params_sum,
                                                      num_examples_sum)
    server_state = server_update(server_state, mean_delta_params)
    return server_state, client_diagnostics

  def server_update(server_state, mean_delta_params):
    opt_state, params = server_optimizer.apply(mean_delta_params,
                                               server_state.opt_state,
                                               server_state.params)
    return ServerState(params, opt_state)

  return federated_algorithm.FederatedAlgorithm(init, apply)
